#!/usr/local/bin/perl
# wsh-c - CGI based remote unix shell (client part)
# by Alex Dyatlov <avd@mts.ru>
# September, 2002
#
# INSTALL
# Module Term::ReadLine::Gnu installation is recommended, get:
# 1) readline-4.2a.tar.gz or later from
#    http://www.gnu.org/directory/readline.html
# 2) ReadLine-Gnu-1.12.tar.gz or later from
#    http://search.cpan.org/search?dist=Term-ReadLine-Gnu
#
# SHELL COMMANDS
# exit		as is
# history	show commands history
# !<number>	execute command with history <number>
# wshget <file> get <file> from remote host to local directory
# wshput <file> put <file> from local directory to remote host
use strict;

use IO::Socket;
use Term::ReadLine;
use POSIX qw(:sys_wait_h);

#--- config - begin ----------------------------------->8--
my $conf = "wsh-c.conf";	#- config file
my $http_port = 80;		#- default HTTP port
my $proxy_port = 3128;		#- default HTTP proxy port
my $shell_prompt = "wsh#";
my $progres = ".";
my $rand_range = 9999;
my $readkbytes = 100;

my %CONF;
my %HOST;
$CONF{use_proxy} = 0;		#- 0 or 1
$CONF{agent} = "Mozilla/4.0 (compatible; MSIE 6.0; Windows 98)";
$CONF{anticache} = 0;		#- 0 or 1
#--- config - end ------------------------------------->8--

my $VERSION = "1.1.2";
$ENV{PERL_RL} = " o=0"; # use best available ReadLine without ornaments

open(CF, $conf)
	or die "unable to open config file $conf: $!";
while(my $str = <CF>) {
	$str =~ s/(\s+)?#(.*)?//;
	if ($str =~ /(\S+)(\s+)(.+)/) {
		$CONF{$1} = $3;
	} else {
		next;
	}
}
close(CF);

my $request = $ARGV[0];
my $key = $ARGV[1];
die "usage: $0 host[:port]/dir/script KEY"
	unless ($request && $key);

my $hostname;
if ($request =~ /^([^\/]+)(.*)/s) {
	$hostname = $1;
	$request = $2
		if (!$CONF{use_proxy});
	if ($hostname =~ /([^:]+):(.*)/) {
		$hostname = $1;
		$http_port = $2;
	}
} else {
	die "\n* error: unable to parse hostname from $ARGV[0]\n\n";
}
if (!$CONF{use_proxy}) {
	if ($hostname !~ /\d+\.\d+\.\d+\.\d+/) {
		(my $name, my $aliases, my $addrtype, my $length, my @addrs) =
			gethostbyname($hostname)
			or die "unable to resolve hostname '$hostname'\n\n";
		$HOST{ip} = join('.', unpack('C4', $addrs[0]));
	} else {
		$HOST{ip} = $hostname;
	}
	$HOST{port} = $http_port;
} else {
	die "\n* error: proxy_ip is not specified in $conf"
		unless ($CONF{proxy_ip});
	$HOST{ip} = $CONF{proxy_ip};
	if ($CONF{port}) {
		$HOST{port} = $CONF{port};
	} else {
		$HOST{port} = $proxy_port;
	}
	$request = "http://".$request;
}

my $term = Term::ReadLine->new("wsh");
my $OUT = $term->OUT() || *STDOUT;
my $io = undef;
my $file;
my $size = 0;
my $cmd = "pwd";
my $slash = "/";
my $amp = ";";
my @h_list = my @msg = ();
SendViaHTTP();
$msg[1] =~ s/\s*$//s;
my $pwd = $msg[1];
while (defined($cmd = $term->readline("$shell_prompt "))) {
	next if (length($cmd) == 0);
# wsh commands --- begin
	exit 0 if ($cmd =~ /^exit$/s);
	if ($cmd =~ /^history$/s) {
		my $h_counter = 1;
		foreach (@h_list) {
			print $OUT "  ".($h_counter++)."\t$_\n";
		}
		next;
	}
	if ($cmd =~ /^\!(\d+)$/s) {
		($1 > 0 && $1 <= scalar(@h_list)) ?
			$cmd = $h_list[$1-1] :
			next;
	}
	$io = undef;
	if ($cmd =~ /^wsh((get)|(put)) ['" ]*?([^'"]+)/s) {
		$io = $1;
		$file = $4;
		$size = 0;
		($pwd =~ /[\/\\]$/s) ?
			$cmd = "wsh$io \"$pwd$file\"" :
			$cmd = "wsh$io \"$pwd$slash$file\"";
		if ($io =~ /put/) {
			unless (open(FH, $file)) {
				print "$file: $!\n";
				next;
			}
			close(FH);
			$size = -s $file;
		}
	}
# wsh commands --- end
	push(@h_list, $cmd);
	if ($cmd =~ /^cd ['"]?([^'";]+)$/s) {
		my $dir = $1;
		if ($dir !~ /^[\/\\]/s) {
			$pwd = join($slash, ($pwd, $dir));
			$pwd =~ s/[^\/\\]+$slash\.\.//g;
			$pwd =~ s/^[\/\\]\.\./$slash/s;
			$pwd =~ s/[\/\\]{2,}/$slash/g;
			$pwd =~ s/[\/\\]?$//s;
		} else {
			$pwd = $dir;
		}
		next;
	}
	unless (defined($io)) {
		$cmd = "if [ -d '$pwd' ]$amp then cd '$pwd'$amp".
			"else echo 'cd: $pwd: No such file or ".
			"directory'$amp exit 0$amp fi$amp $cmd";
	}
	next
		if (!SendViaHTTP());
	print $OUT $msg[1];
}

sub SendViaHTTP() {
	my $req;
	$cmd =~ s/\s*$//s;
	my $conl;
	if ($io =~ /((get)|(put))/) {
		$conl = $size;
	} else {
		$conl = length($cmd);
	}
	my $SOCKET = IO::Socket::INET->new(
		PeerAddr => $HOST{ip},
		PeerPort => $HOST{port},
		Proto    => "tcp",
		Type     => SOCK_STREAM)
		or die "\n* error: fail to connect to $HOST{ip}:$HOST{port}: $!";
	($CONF{anticache}) ?
		$req = "POST $request?".(int(rand($rand_range)))." HTTP/1.0" :
		$req = "POST $request HTTP/1.0";
	$req = join("\r\n", $req,
		"Content-Type: application/octet-stream",
		"User-Agent: $CONF{agent}",
		"Host: $hostname",
		"Content-Length: $conl",
		"X-Key: $key"
	);
	($CONF{use_proxy}) ?
		$req = join("\r\n", $req,	
			"Proxy-Connection: close",
			"Pragma: no-cache") :
		$req = join("\r\n", $req,	
			"Connection: close");
	if ($io =~ /put/ || $io =~/get/) {
		$req = join("\r\n", $req,	
			"X-File$io: $pwd$slash$file",
			"\r\n");
	} else {
		$req = join("\r\n\r\n", $req,
			$cmd);
	}
	print $SOCKET $req;
	if ($io =~ /put/) {
		unless (open(FH, $file)) {
			print "$file: $!\n";
			return 0;
		}
		while ((my $bytes = read(FH, my $buf,
			$readkbytes * 1024,
			my $position))) {
			$position += $bytes;
			while ($bytes) {
				unless ($b = syswrite($SOCKET, $buf, $bytes)) {
					print $OUT "\n* error: fail to send file $file\n";
					return 0;
				}
				$buf = substr($buf, $b, $bytes - $b)
					if ($bytes != $b);
				$bytes -= $b;
			}
			syswrite($OUT, $progres, length($progres));
		}
		print $OUT "\n";
		close(FH);
	}
	my $cl = 0;
	@msg = ();
	while (my $str = <$SOCKET>) {
		last
			if ($str =~ /^\s*?$/s);
		$msg[0] = $msg[0].$str;
	}
	$msg[0] =~ /^[^ ]+ 200/s
		or die "\n* error: host $HOST{ip} answer:\n$msg[0]\n";
	if ($io =~ /get/) {
		unless (open(FH, "> $file")) {
			print $OUT "\n* error: fail to get file $file: $!\n";
			close($SOCKET);
			return 0;
		}
		while (my $bytes = read($SOCKET, my $buf, $readkbytes * 1024)) {
			my $b = 0;
			while ($bytes) {
				unless ($b = syswrite(FH, $buf, $bytes)) {
					print $OUT "\n* error: fail to get file $file\n";
					return 0;
				}
				$bytes -= $b;
			}
			syswrite($OUT, $progres, length($progres));
		}
		print $OUT "\n";
		close(FH);
		close($SOCKET);
		return 1;
	}
	$cl = length($msg[0]) + 4;
	if ($msg[0] =~ /Content-Length: (\d+)/s) {
		$cl += $1;
	} else {
		$cl = -1;
	}
	while (my $str = <$SOCKET>) {
		$msg[1] = $msg[1].$str;
		last if ($cl > 0 && length($msg[0].$msg[1])+4 >= $cl);
	}
	close($SOCKET);
	return 1;
}
